
%% Initializing lambdaa, mue and sos variables
% Lambda
lambdae = sdpvar(nx,1);
lambdau = sdpvar(nuc,1);
lambdaq = sdpvar(1,1);
% Mu
mu1 = sdpvar(nu,nx);
mu2 = sdpvar(nx,nx);
mu4 = sdpvar(nx,1);
c1 = sdpvar(1,1);
% Polynomials      % SOS polynomials in x of degrees 4
[s4,cc4,v4] = polynomial([x;xp],s4_deg);
[s5,cc5,v5] = polynomial([x;xp],s5_deg);
[lambdaq,lambdaq_c, lambdaq_v] = polynomial ([x;xp],lambdaq_deg);

mu1c = cell(nu,nx);

for j = 1:nu
    for k = 1:nx
        if ((j == 1) && (k == 1))
            [mu1(j,k),mu1c{j,k},mu1v] = polynomial([x;xp],mu1_deg);
        else 
            [mu1(j,k),mu1c{j,k}] = polynomial([x;xp],mu1_deg);
        end
    end
end

mu2c = cell(nx,nx);

for j = 1:nx
    for k = 1:nx
        if ((j == 1) && (k == 1))
            [mu2(j,k),mu2c{j,k},mu2v] = polynomial([x;xp],mu2_deg);
        else 
            [mu2(j,k),mu2c{j,k}] = polynomial([x;xp],mu2_deg);
        end
    end
end

s3 = sdpvar(nuc,1);
for j=1:nuc
    [s3(j),cc3{j},v3{j}] = polynomial(x,s3_deg);
    [lambdau(j),lambdau_c{j}, lambdau_v{j}] = polynomial (x,lambdau_deg);  
end

for j=1:nx
    [mu4(j),mu4c{j},mu4v{j}] = polynomial([x;xp],mu4_deg);
    [lambdae(j),lambdae_c{j}, lambdae_v{j}] = polynomial ([x;xp],lambdae_deg);  
end

%Region of interest
px = -(1-Vx)*(1-x'*P0*x);       
% Degree of Vxp should be same as of Vx
degre = degree(Vx);
if (length(Vxc) == 42)
    degre = 8;
end
% Vxp is a polynomial of xp, which is the next state
[Vxp,Vpc,Vpv] = polynomial(xp,degre,2);
% Coefficients of Vxp should be same as Vx.
len = size(Vpv,1);
Vxc(len+1:end)=[];
Vxp = Vxc'*Vpv;


%% Constraints
% Diff. of lagrangian wrt u
E1 = R*u-B'*lambdae+Hu'*lambdau+mu1*(xp-A*x-B*u);
% Diff. of lagrangian wrt x_{k+1}
E2 = Q*xp+lambdae+(lambdaq*jacobian(Vxp,xp))'+mu2*(xp-A*x-B*u);
%% Input Constraints
%F1 = -Hu*u+hu-s1*px;
%% Contractive Constraint
%F2 = -Vxp + gamma*Vx - mu3'*(xp-A*x-B*u)- s2*px;
% Lagrange Multipliers
F3 = lambdau - s3*px;
F4 = lambdaq - s4*px;
F5 = lambdau'*(Hu*u-hu)+lambdaq*(Vxp -gamma*Vx) -mu4'*(xp-A*x-B*u) - s5*px+c1*(x'*x);

Model = [];

% All Constraints are added to Model
Model = [Model, coefficients(E1,[x,xp])==0,coefficients(E2,[x,xp])==0,sos(F3),sos(F4),sos(F5),...
    sos(s3),sos(s4),sos(s5),...
    c1>=0,c1<=0.05];

% Parameters
params = [cc4;cc5;lambdaq_c;c1];  %Parameters of s4,s5,s6 and lambda_c

%Parameters of mu1
for j = 1:nu
    for k = 1:nx
        params = [params;mu1c{j,k}];
    end
end

%Parameters of mu2
for j = 1:nx
    for k = 1:nx
        params = [params;mu2c{j,k}];
    end
end

%Parameters of s3 and lambda_u
for j=1:nuc
    params = [params;cc3{j};lambdau_c{j}];
end


% Parameters of lambdae and mu4
for j=1:nx
    params = [params;lambdae_c{j};mu4c{j}];
end

optacc = 1e-10;

% Solving SOS
% ops =sdpsettings('savesolveroutput',1);
%ops = sdpsettings('sos.model',1,'sos.postprocess',1,'sos.numblkdg',1e-10,'solver','sdpt3','sdpt3.gaptol',1e-10,'sdpt3.inftol',1e-10,'sdpt3.steptol',1e-10,'sdpt3.maxit',100);
ops = sdpsettings('solver','mosek','mosek.MSK_DPAR_INTPNT_CO_TOL_DFEAS',optacc,'mosek.MSK_DPAR_INTPNT_CO_TOL_PFEAS',optacc,'mosek.MSK_DPAR_INTPNT_CO_TOL_REL_GAP',optacc,'sos.numblkdg',1e-10,'sos.model',1,'sos.postprocess',1);
%ops = sdpsettings(ops,'mosek.MSK_IPAR_PRESOLVE_USE','MSK_PRESOLVE_MODE_OFF','mosek.MSK_IPAR_INTPNT_SCALING','MSK_SCALING_NONE');
%[sol,v,Qs,res] =solvesos(Model,c1,ops,params);
[sol,v,Qs,res] =solvesos(Model,[],ops,params);
sprob_lambda = sol.problem
if (max(res) > 1e-6)
    disp('No solution found in lambda, trying dual')
    sprob_lambda = -10
end

if (sprob_lambda ~= 0)
    %ops = sdpsettings('sos.model',2,'sos.postprocess',0,'sos.numblkdg',0,'solver','sdpt3','sdpt3.gaptol',1e-10,'sdpt3.inftol',1e-10,'sdpt3.steptol',1e-10,'sdpt3.maxit',100);
    ops = sdpsettings('solver','mosek','mosek.MSK_DPAR_INTPNT_CO_TOL_DFEAS',optacc,'mosek.MSK_DPAR_INTPNT_CO_TOL_PFEAS',optacc,'mosek.MSK_DPAR_INTPNT_CO_TOL_REL_GAP',optacc,'sos.numblkdg',0,'sos.model',2,'sos.postprocess',0);
  %  ops = sdpsettings(ops,'mosek.MSK_IPAR_PRESOLVE_USE','MSK_PRESOLVE_MODE_OFF','mosek.MSK_IPAR_INTPNT_SCALING','MSK_SCALING_NONE');
 %   [sol,v,Qs,res] =solvesos(Model,c1,ops,params);
    [sol,v,Qs,res] =solvesos(Model,[],ops,params);
    sprob_lambda = sol.problem
    maxres = max(res)
end

if (max(res) > 1e-6)
    disp('No solution found in lambda')
    sprob_lambda = -10
end

if (sprob_lambda == 0)
%% Obtaining Solution
    S2 = value(cc2)'*v2;
    S4 = value(cc4)'*v4;
    S5 = value(cc5)'*v5;
    Lambdaq = value(lambdaq_c)'*lambdaq_v;
    Mu2 = sdpvar(nx,nx);
    for j=1:nx
        for k= 1:nx
            Mu2(j,k) = value(mu2c{j,k})'*mu2v;
        end
    end
    Mu1 = sdpvar(nu,nx);
    for j=1:nu
        for k= 1:nx
            Mu1(j,k) = value(mu1c{j,k})'*mu1v;
        end
    end
   
    
    for j = 1:nuc
        Lambdau(j,1) = value(lambdau_c{j})'*lambdau_v{j};
        S3(j,1) = value(cc3{j})'*v3{j};
    end

    for j=1:nx
        Lambdae(j) = value(lambdae_c{j})'*lambdae_v{j};
        Mu4(j,1) = value(mu4c{j})'*mu4v{j};
    end
    
end

% Checking SOS decomposition
len=size(res,1);
for j=1:len
    remainder(j) = min(eig(Qs{j})) - len*res(j); 
end
val1=value(c1)
