
% Input
u = sdpvar(nu,1);
z = sdpvar(nx,1);
lambdae = sdpvar(nx,1);
c1 = sdpvar(1,1);

Xpoly = Polyhedron(Hxp,ones(size(Hxp,1),1));
% Vx is a lyapunov function Vx <=1 with minimum degree of 2
[Vx,Vxc,Vxv] = polynomial(x,deg,2);
%Region of interest
px = -(1-Vx)*(1-x'*P0*x);       
[s6,cc6,v6] = polynomial(x,s6_deg);
for j=1:nx
    [lambdae(j),lambdae_c{j},lambdae_v{j}] = polynomial([x;xp],lambdae_deg);
end
for j=1:nu
    [u(j),u_c{j}, u_v{j}] = polynomial (x,udeg,1); 
end
[Vxp,Vpc,Vpv] = polynomial(xp,deg,2);
% Coefficients of Vxp should be same as Vx.f
Vxp = Vxc'*Vpv;

% Points on Omega touching the state constraints:
xx1pos = [5 0];
xx1neg = [-5 0];
Vxx1neg = replace(Vx, x, xx1neg);
Vxx1pos = replace(Vx, x, xx1pos);
xx2pos = [0 5];
xx2neg = [0 -5];
Vxx2neg = replace(Vx, x, xx2neg);
Vxx2pos = replace(Vx, x, xx2pos);
%sdisplay(Vxx1pos)
%pause()

%% Constraints
% Diff. of lagrangian wrt u
E1 = R*u-B'*lambdae+Hu'*Lambdau+Mu1*(xp-A*x-B*u);
% Diff. of lagrangian wrt x_{k+1}
E2 = Q*xp+lambdae+(Lambdaq*jacobian(Vxp,xp))'+Mu2*(xp-A*x-B*u);
%Hard constraint on breaking state constraints
E3 = Vxx1pos-1;
E4 = Vxx1neg-1;
% Input Constraints
F1 = -Hu*u+hu-S1*px;
% Contractive Constraint
F2 = -Vxp + gamma*Vx - Mu3*(xp-A*x-B*u)- S2*px;
% Lagrange Multipliers
F3 = Lambdau - S3*px;
F4 = Lambdaq - S4*px;
F5 = Lambdau'*(Hu*u-hu)+Lambdaq*(Vxp -gamma*Vx) -Mu4'*(xp-A*x-B*u) - S5*px+c1*(x'*x);
% Vx<=1 should be greater than Vxprev<=1
F6 = -(Vx-1)-s6*(1-Vxprev);
%F6 = -(Vx-1)-s6*(1-1.001*Vxprev); %Cheating 
d2dxV = hessian(Vx,x);
%sdisplay(hess)
F7 = z'*d2dxV*z;
%F7 = hessian(Vx)-s8*px;
Model = [];
% State Constraints
for i=1:nx
    Model = [Model,sos((-(Hxp(i,:)*x-1)-S7(i)*(1-Vx)))];
    %Model = [Model,sos((-(Hxp(i,:)*x-1)-S7(i)*(1-1.001*Vx)))]; %Cheating
end

% Point Constraint - The point lies inside polyhedral contractive set.
% val and vals should be less than or equal to one in order to point 
% and-point lie within Vx.
%Vxx1neg<=1,Vxx1pos<=1,Vxx1neg>=0.99,Vxx1pos>=0.99
val = replace(Vx,x,point);
vals = replace(Vx,x,-point);
% All Constraints are added to Model
Model = [Model,coefficients(E1,[x,xp])==0,coefficients(E2,[x,xp])==0,sos(F1),sos(F2),sos(F3),sos(F4),sos(F5),sos(F6),sos(F7),...
    sos(s6),sos(Vx),val<=1,vals<=1,...Vxx1neg>=0.99,Vxx1neg<=1,Vxx1pos>=0.99,Vxx1pos<=1,Vxx2neg>=0.99,Vxx2neg<=1,Vxx2pos>=0.99,Vxx2pos<=1,...
    c1>=0,c1 <=0.05];

% Parameters
% Parameters for Vx and s6
params=[Vxc;cc6;c1];
% Parameters of lambdae
for i=1:nx
    params = [params;lambdae_c{i}];
end
% Parameters for u
for i=1:nu
    params = [params;u_c{i}];
end
disp('Optimizing for u and v')
% Solver options
% ops =sdpsettings('savesolveroutput',1);
%ops = sdpsettings('sos.model',1,'sos.postprocess',1,'sos.numblkdg',1e-10,'solver','sdpt3','sdpt3.gaptol',1e-10,'sdpt3.inftol',1e-10,'sdpt3.steptol',1e-10,'sdpt3.maxit',100);
ops = sdpsettings('solver','mosek','mosek.MSK_DPAR_INTPNT_CO_TOL_DFEAS',1e-10,'mosek.MSK_DPAR_INTPNT_CO_TOL_PFEAS',1e-10,'mosek.MSK_DPAR_INTPNT_CO_TOL_REL_GAP',1e-10,'sos.congruence',1,'sos.numblkdg',1e-12,'sos.model',1,'sos.postprocess',1);
%ops = sdpsettings(ops,'solver','mosek','mosek.MSK_DPAR_INTPNT_CO_TOL_DFEAS',1e-12,'mosek.MSK_DPAR_INTPNT_CO_TOL_PFEAS',1e-12,'mosek.MSK_DPAR_INTPNT_CO_TOL_REL_GAP',1e-12,'mosek.MSK_IPAR_NUM_THREADS',8,'sos.congruence',0,'sos.numblkdg',0,'sos.model',2,'sos.postprocess',1);

%ops = sdpsettings(ops,'mosek.MSK_IPAR_PRESOLVE_USE','MSK_PRESOLVE_MODE_OFF','mosek.MSK_IPAR_INTPNT_SCALING','MSK_SCALING_NONE');
% ops = sdpsettings(ops,'mosek.MSK_IPAR_INTPNT_SOLVE_FORM','MSK_SOLVE_DUAL');
%[sol,v,Qs,res] =solvesos(Model,c1,ops,params);
[sol,v,Qs,res] =solvesos(Model,[],ops,params);
sprob_uv = sol.problem
if (max(res) > 1e-6)
    sprob_uv = -10
end

if (sprob_uv ~= 0)
    %ops = sdpsettings('sos.model',2,'sos.postprocess',0,'sos.numblkdg',0,'solver','sdpt3','sdpt3.gaptol',1e-10,'sdpt3.inftol',1e-10,'sdpt3.steptol',1e-10,'sdpt3.maxit',100);
    ops = sdpsettings('solver','mosek','mosek.MSK_DPAR_INTPNT_CO_TOL_DFEAS',1e-10,'mosek.MSK_DPAR_INTPNT_CO_TOL_PFEAS',1e-10,'mosek.MSK_DPAR_INTPNT_CO_TOL_REL_GAP',1e-10,'sos.congruence',0,'sos.numblkdg',0,'sos.model',2,'sos.postprocess',0);
    %ops = sdpsettings(ops,'mosek.MSK_IPAR_PRESOLVE_USE','MSK_PRESOLVE_MODE_OFF','mosek.MSK_IPAR_INTPNT_SCALING','MSK_SCALING_NONE');
% ops = sdpsettings(ops,'mosek.MSK_IPAR_INTPNT_SOLVE_FORM','MSK_SOLVE_DUAL');
 %   [sol,v,Qs,res] =solvesos(Model,c1,ops,params);
    [sol,v,Qs,res] =solvesos(Model,[],ops,params);
    sprob_uv = sol.problem
    maxres = max(res)
end

if (max(res) > 1e-6)
    sprob_uv = -10
end

if (sprob_uv == 0)
    % Obtaining solution
    Vxc = value(Vxc);
    Vx = Vxc'*Vxv;

    % Obtaining u
    for j=1:nu
        u(j) = value(u_c{j})'* u_v{j};  
        uc = value(u_c{j});  % For single input
        Ku(j,:) = value(u_c{j})';
    end

%s6 = value(cc6)'*v6; %Only used here - not needed anywhere else
end


%% Checks

assign(x,point)
assign(xp,value(A*x+B*u))
disp('F1 is:')
value(F1)  % Input Constraint
disp('F2 is:')
value(F2) % Contraction COnstraint
disp('State constraint is:')
value(-(Hxp(1,:)*x-1)-S7(1)*(1-Vx))
disp('u is:')
value(u)
disp('Degree of Vx is:')
degree(Vx)
disp('Degree of u is:')
degree(u)
% str1 = 'MSK_RES_TRM_STALL';
% str2 = sol.solveroutput.res.rcodestr;
% check = strcmp(str1,str2);
% if value(F5) < 0 || sol.problem~=0
% Vx = Vxprev;
% err = 1;
% else
%     Vxc = value(Vxc);
%len=size(res,1);
%    for j=1:len
%         lenQ = size(Qs{j},1);
%         remainder(j) = min(eig(Qs{j})) - lenQ*res(j); 
%    end
%val1=value(c1);
% end


